/*
 settings.json:
    "markdown-pdf.styles": ["markdown-pdf.css",],
    "markdown-pdf.format": "Letter",
    "markdown-pdf.margin.top": "1in",
    "markdown-pdf.margin.bottom": "1in",
    "markdown-pdf.margin.left": "1in",
    "markdown-pdf.margin.right": "1in",
    "markdown-pdf.stylesRelativePathFile" : true,
    "markdown-pdf.displayHeaderFooter": false,
 */

body {
  font-family: "Times";
  font-size: 10pt;
  padding: 0;
}

h4 {
  font-family: "Times New Roman";
  font-size: 10pt;
  font-weight: bold;
}

code {
  font-family: Consolas, "Source Code Pro";
  font-size: 10pt;
}

pre.hljs code > div {
  padding: 0px;
}

:not(pre):not(.hljs) > code {
  color: #4d4d4c;
}
