// Copyright (c) the JPEG XL Project Authors. All rights reserved.
//
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

namespace jxl {
void BenchmarkFloatIDCT32x32();
void BenchmarkFastIDCT32x32();
}  // namespace jxl
